# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import datapackage
from .inspector import Inspector


# Module API

def validate(source, **options):
    """Validates a source file and returns a report.

    Args:
        source (Union[str, Dict, List[Dict], IO]): The source to be validated.
            It can be a local file path, URL, dict, list of dicts, or a
            file-like object. If it's a list of dicts and the `preset` is
            "nested", each of the dict key's will be used as if it was passed
            as a keyword argument to this method.

            The file can be a CSV, XLS, JSON, and any other format supported by
            `tabulator`_.

    Keyword Args:
        checks (List[str]): List of checks names to be enabled. They can be
            individual check names (e.g. `blank-headers`), or check types (e.g.
            `structure`).
        skip_checks (List[str]): List of checks names to be skipped. They can
            be individual check names (e.g. `blank-headers`), or check types
            (e.g.  `structure`).
        infer_schema (bool): Infer schema if one wasn't passed as an argument.
        infer_fields (bool): Infer schema for columns not present in the received schema.
        order_fields (bool): Order source columns based on schema fields order.
            This is useful when you don't want to validate that the data
            columns' order is the same as the schema's.
        error_limit (int): Stop validation if the number of errors per table
            exceeds this value.
        table_limit (int): Maximum number of tables to validate.
        row_limit (int): Maximum number of rows to validate.

        preset (str): Dataset type could be `table` (default), `datapackage`,
            `nested` or custom. Usually, the preset can be inferred from the
            source, so you don't need to define it.
        Any (Any): Any additional arguments not defined here will be passed on,
            depending on the chosen `preset`. If the `preset` is `table`, the
            extra arguments will be passed on to `tabulator`_, if it is
            `datapackage`, they will be passed on to the `datapackage`_
            constructor.

        # Table preset
        schema (Union[str, Dict, IO]): The Table Schema for the
            source.
        headers (Union[int, List[str]): Either the row number that contains
            the headers, or a list with them. If the row number is given, ?????
        scheme (str): The scheme used to access the source (e.g. `file`,
            `http`). This is usually inferred correctly from the source. See
            the `tabulator`_ documentation for the list of supported schemes.
        format (str): Format of the source data (`csv`, `datapackage`, ...).
            This is usually inferred correctly from the source. See the
            the `tabulator`_ documentation for the list of supported formats.
        encoding (str): Encoding of the source.
        skip_rows (Union[int, List[Union[int, str]]]): Row numbers or a
            string. Rows beginning with the string will be ignored (e.g. '#',
            '//').

    Raises:
        GoodtablesException: Raised on any non-tabular error.

    Returns:
        dict: The validation report.

    .. _tabulator:
        https://github.com/frictionlessdata/tabulator-py
    .. _tabulator_schemes:
        https://github.com/frictionlessdata/tabulator-py
    .. _tabulator:
        https://github.com/frictionlessdata/datapackage-py
    """
    source, options, inspector_settings = _parse_arguments(source, **options)

    # Validate
    inspector = Inspector(**inspector_settings)
    report = inspector.inspect(source, **options)

    return report


def init_datapackage(resource_paths):
    """Create tabular data package with resources.

    It will also infer the tabular resources' schemas.

    Args:
        resource_paths (List[str]): Paths to the data package resources.

    Returns:
        datapackage.Package: The data package.
    """
    dp = datapackage.Package({
        'name': 'change-me',
        'schema': 'tabular-data-package',
    })

    for path in resource_paths:
        dp.infer(path)

    return dp


def _parse_arguments(source, **options):
    # Extract settings
    validation_options = set((
        'checks',
        'skip_checks',
        'infer_schema',
        'infer_fields',
        'order_fields',
        'error_limit',
        'table_limit',
        'row_limit',
        'custom_presets',
        'custom_checks',
    ))
    settings = dict((
        (key, options.pop(key)) for key in validation_options
        if key in options
    ))

    # Support for pathlib.Path
    if hasattr(source, 'joinpath'):
        source = str(source)
    if isinstance(source, list):
        if source and isinstance(source[0], dict) and 'source' in source[0]:
            for index, item in enumerate(source):
                if hasattr(item['source'], 'joinpath'):
                    source[index]['source'] = str(item['source'])

    return source, options, settings



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import io
import os
import json


# Internal

def _load_spec():
    path = os.path.join(os.path.dirname(__file__), 'spec.json')
    spec = json.load(io.open(path, encoding='utf-8'))
    return spec


# Module API

spec = _load_spec()



from six.moves import zip_longest


def create_cells(headers, schema_fields, values=None, row_number=None):
    """Create list of cells from headers, fields and values.

    Args:
        headers (List[str]): The headers values.
        schema_fields (List[tableschema.field.Field]): The tableschema
            fields.
        values (List[Any], optional): The cells values. If not specified,
            the created cells will have the same values as their
            corresponding headers. This is useful for specifying headers
            cells.
            If the list has any `None` values, as is the case on empty
            cells, the resulting Cell will have an empty string value. If
            the `values` list has a different length than the `headers`,
            the resulting Cell will have value `None`.
        row_number (int, optional): The row number.

    Returns:
        List[dict]: List of cells.
    """
    fillvalue = '_fillvalue'
    is_header_row = (values is None)
    cells = []

    iterator = zip_longest(headers, schema_fields, values or [], fillvalue=fillvalue)
    for column_number, (header, field, value) in enumerate(iterator, start=1):
        if header == fillvalue:
            header = None
        elif is_header_row:
            value = header
        if field == fillvalue:
            field = None
        if value == fillvalue:
            value = None
        elif value is None:
            value = ''

        cell = create_cell(header, value, field, column_number, row_number)
        cells.append(cell)

    return cells


def create_cell(header, value=None, field=None, column_number=None, row_number=None):
    cell = {
        'header': header,
        'field': field,
        'value': value,
        'column-number': column_number,
        'number': column_number,  # FIXME: Remove deprecated "number"
        'row-number': row_number,
    }
    return _remove_none_values(cell)


def _remove_none_values(dictionary):
    result = {}
    for key, value in dictionary.items():
        if value is not None:
            result[key] = value
    return result



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import six
from tabulator import Stream
from tableschema import Schema, exceptions
from ..registry import preset


# Module API

@preset('table')
def table(source, schema=None, **options):
    warnings = []
    tables = []

    # Ensure not a datapackage
    if isinstance(source, six.string_types):
        if source.endswith('datapackage.json'):
            warnings.append('Use "datapackage" preset for Data Packages')

    # Prepare schema
    if not warnings:
        if schema is not None:
            try:
                schema = Schema(schema)
            except exceptions.TableSchemaException as error:
                warnings.append(
                    'Table Schema "%s" has a loading error "%s"' %
                    (schema, error))

    # Add table
    if not warnings:
        options.setdefault('headers', 1)
        tables.append({
            'source': str(source) if isinstance(source, six.string_types) else 'inline',
            'stream': Stream(source, **options),
            'schema': schema,
            'extra': {},
        })

    return warnings, tables



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..registry import check
from ..error import Error


# Module API

@check('extra-value')
class ExtraValue(object):
    def __init__(self, **options):
        self._num_columns = None

    def check_row(self, cells):
        errors = []

        # Check that all rows have the same number of columns
        if self._num_columns is None:
            self._num_columns = len(cells)
        elif len(cells) > self._num_columns:
            extra_cells = cells[self._num_columns:]
            for cell in extra_cells:
                error = Error('extra-value', cell)
                errors.append(error)
                cells.remove(cell)
            return errors

        # Check that if any cell has a header, all cells MUST have headers as
        # well. This deals with the case of data files without headers.
        headers = [cell.get('header') for cell in cells]
        has_header = any(headers)
        cells_without_header = filter(lambda cell: cell.get('header') is None, cells)
        if has_header and cells_without_header:
            for cell in cells_without_header:
                error = Error('extra-value', cell)
                errors.append(error)
                cells.remove(cell)

        return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
from ..registry import check
from ..error import Error


# Module API

@check('required-constraint')
def required_constraint(cells):
    errors = []

    for cell in copy(cells):

        field = cell.get('field')
        value = cell.get('value')

        # Skip if cell has no field
        if field is None:
            continue

        # Check constraint
        valid = field.test_value(value, constraints=['required'])
        if field.descriptor.get('primaryKey'):
            valid = valid and field.cast_value(value) is not None

        # Skip if valid
        if valid:
            continue

        # Add error

        error = Error('required-constraint', cell)
        errors.append(error)

        # Remove cell
        cells.remove(cell)

    return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..error import Error


def create_check_constraint(check, constraint):
    def _check_constraint(cells):
        errors = []

        for cell in cells:

            field = cell.get('field')
            value = cell.get('value')

            # Skip if cell has no field
            if field is None:
                continue

            # Check constraint
            valid = field.test_value(value, constraints=[constraint])

            # Add error
            if not valid:
                message_substitutions = {
                    'value': '"{}"'.format(value),
                    'constraint': '"{}"'.format(field.constraints[constraint]),
                }

                error = Error(
                    check,
                    cell,
                    message_substitutions=message_substitutions
                )
                errors.append(error)

        return errors

    return _check_constraint



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from goodtables.presets.table import table
from goodtables.presets.datapackage import datapackage
from goodtables.presets.nested import nested


# Preset

def test_preset_nested():
    warnings, tables = nested([
        {'source': 'data/valid.csv'},
        {'source': 'data/invalid.csv', 'preset': 'table'},
        {'source': 'data/datapackages/valid/datapackage.json', 'preset': 'datapackage'},
        {'source': 'data/datapackages/invalid/datapackage.json', 'preset': 'datapackage'},
    ], presets={
        'table': {'func': table},
        'nested': {'func': nested},
        'datapackage': {'func': datapackage},
    })
    assert len(warnings) == 0
    assert len(tables) == 6



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals


# Base

class GoodtablesException(Exception):
    """https://github.com/frictionlessdata/goodtables-py#exceptions
    """
    pass



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ...registry import check
from ...error import Error


# Module API

@check('sequential-value', type='custom', context='body')
class SequentialValue(object):

    # Public

    def __init__(self, column, **options):
        self.__column = column
        self.__cursor = None
        self.__code = 'sequential-value'

    def check_row(self, cells):

        # Get cell
        cell = None
        for item in cells:
            if self.__column in [item['column-number'], item['header']]:
                cell = item
                break

        # Check cell
        if not cell:
            message = 'Sequential value check requires column "{column}" to exist'
            message = message.format(column=self.__column)
            error = Error(
                self.__code,
                row_number=cells[0]['row-number'],
                message=message
            )
            return [error]

        # Get value
        try:
            value = int(cell.get('value'))
        except (TypeError, ValueError):
            message = (
                'Sequential value check requires column "{column_number}"'
                ' to be an integer'
            )
            error = self._create_error(cell, message)
            return [error]

        # Initiate cursor
        if self.__cursor is None:
            self.__cursor = value

        # Check value
        if self.__cursor != value:
            self.__cursor = value + 1
            message = (
                'Value "{value}" is not a sequential in column {column_number}'
                ' for row {row_number}'
            )
            message_substitutions = {
                'value': value,
            }
            error = self._create_error(cell, message, message_substitutions)
            return [error]

        # Increment cursor
        self.__cursor += 1

    def _create_error(self, cell, message, message_substitutions=None):
        return Error(
            self.__code,
            cell,
            message=message,
            message_substitutions=message_substitutions
        )



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import statistics
from ...registry import check
from ...error import Error
from ... import exceptions


# Module API

@check('deviated-value', type='custom', context='body')
class DeviatedValue(object):

    # Public

    def __init__(self, column, average='mean', interval=3, **options):

        # Set attributes
        self.__column = column
        self.__interval = interval
        self.__column_cells = []
        self.__average_function = _AVERAGE_FUNCTIONS.get(average)
        self.__code = 'deviated-value'
        self.__cell = None

        # Validate average
        if not self.__average_function:
            message = 'Deviated value check supports only this average functions: %s'
            message = message % ', '.join(_AVERAGE_FUNCTIONS.keys())
            raise exceptions.GoodtablesError(message)

    def check_row(self, cells):

        # Get cell
        for cell in cells:
            if self.__column in [cell['column-number'], cell['header']]:
                self.__cell = cell
                break

        # Check cell
        if not self.__cell:
            message = 'Deviated value check requires column "{column}" to exist'
            message = message.format(column=self.__column)
            error = Error(
                self.__code,
                row_number=cells[0]['row-number'],
                message=message
            )
            return [error]

        # Get value
        try:
            float(cell['value'])
        except ValueError:
            message = (
                'Deviated value check requires column "{column_number}"'
                ' to be a number'
            )
            error = Error(
                self.__code,
                self.__cell,
                message=message
            )
            return [error]

        # Collect value
        self.__column_cells.append(cell)

    def check_table(self):

        # Skip if not values
        if not self.__column_cells or len(self.__column_cells) < 2:
            return

        # Prepare interval
        try:
            values = [float(cell['value']) for cell in self.__column_cells]
            stdev = statistics.stdev(values)
            average = self.__average_function(values)
            minimum = average - stdev * self.__interval
            maximum = average + stdev * self.__interval
        except Exception as exception:
            message = 'Deviated value check calculation issue: %s' % exception
            error = Error(
                self.__code,
                self.__cell,
                message=message
            )
            return [error]

        # Check values
        errors = []
        for cell in self.__column_cells:
            if not (minimum <= float(cell['value']) <= maximum):
                message = (
                    'Deviated value "{value}" in column {column_number}'
                    ' for row {row_number}'
                )
                message_substitutions = {
                    'value': cell['value'],
                }
                error = Error(
                    self.__code,
                    cell,
                    message=message,
                    message_substitutions=message_substitutions
                )
                errors.append(error)

        return errors


# Internal

_AVERAGE_FUNCTIONS = {
    'mean': statistics.mean,
    'median': statistics.median,
    'mode': statistics.mode,
}



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import copy
import tableschema
from ..registry import check
from ..error import Error


# Module API

@check('type-or-format-error')
def type_or_format_error(cells):
    errors = []

    for cell in copy(cells):

        field = cell.get('field')
        value = cell.get('value')

        # Skip if cell has no field
        if field is None:
            continue

        # Cast value
        try:
            valid = True
            field.cast_value(value, constraints=False)
        except tableschema.exceptions.TableSchemaException:
            valid = False

        # Skip if valid
        if valid:
            continue

        # Add error
        message_substitutions = {
            'value': '"{}"'.format(value),
            'field_type': '"{}"'.format(field.type),
            'field_format': '"{}"'.format(field.format),
        }
        error = Error(
            'type-or-format-error',
            cell,
            message_substitutions=message_substitutions
        )
        errors.append(error)

        # Remove cell
        cells.remove(cell)

    return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from goodtables.presets.table import table


# Preset

def test_preset_table():
    warnings, tables = table('data/valid.csv')
    assert len(warnings) == 0
    assert len(tables) == 1


def test_preset_table_but_got_datapackage_issue_187():
    warnings, tables = table('data/datapackages/valid/datapackage.json')
    assert len(warnings) == 1
    assert len(tables) == 0
    assert 'Use "datapackage" preset' in warnings[0]


def test_preset_table_invalid_json_issue_196():
    warnings, tables = table('valid.csv', schema='data/invalid_json.json')
    assert len(warnings) == 1
    assert len(tables) == 0
    assert 'has a loading error' in warnings[0]



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import datetime
import pytest
import goodtables.cells
from goodtables import validate
from goodtables.contrib.checks.truncated_value import TruncatedValue


@pytest.mark.parametrize('value', [
    'a' * 255,
    32767,
    2147483647,
])
def test_truncated_values(value):
    cell = goodtables.cells.create_cell('value', value)
    errors = TruncatedValue().check_row([cell])
    assert len(errors) == 1
    assert errors[0].code == 'truncated-value'


@pytest.mark.parametrize('value', [
    'a' * 254,
    32766,
    2147483646,
])
def test_not_truncated_values(value):
    cell = goodtables.cells.create_cell('value', value)
    errors = TruncatedValue().check_row([cell])
    assert not errors


def test_check_truncated_value_with_dates():
    # There was a bug where we didn't catch the correct exception when calling
    # int(date)
    cell = goodtables.cells.create_cell(
        'date',
        datetime.datetime.now()
    )
    errors = TruncatedValue().check_row([cell])

    assert not errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..registry import check
from .constraints_checks import create_check_constraint


# Module API

@check('pattern-constraint')
def pattern_constraint(cells):
    errors = _check_constraint(cells)

    for error in errors:
        # No reason to run other checks in this cell
        cells.remove(error.cell)

    return errors


_check_constraint = create_check_constraint('pattern-constraint', 'pattern')



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from copy import deepcopy
from ..registry import preset
from .. import exceptions


# Module API

@preset('nested')
def nested(source, presets):
    warnings = []
    tables = []

    # Add warnings, tables
    source = deepcopy(source)
    for item in source:
        preset = item.pop('preset', 'table')
        checks = item.pop('checks', [])
        if preset == 'nested':
            message = 'Preset "nested" supports only one level depth'
            raise exceptions.GoodtablesException(message)
        try:
            preset_func = presets[preset]['func']
        except KeyError:
            message = 'Preset "%s" is not registered' % preset
            raise exceptions.GoodtablesException(message)
        item_warnings, item_tables = preset_func(**item)
        if checks:
            for item_table in item_tables:
                item_table['checks'] = checks
        warnings.extend(item_warnings)
        tables.extend(item_tables)

    return warnings, tables



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..registry import check
from ..error import Error


# Module API

@check('blank-row')
def blank_row(cells):
    errors = []

    if not list(filter(lambda cell: cell.get('value'), cells)):

        # Add error
        error = Error('blank-row', row_number=cells[0].get('row-number'))
        errors.append(error)

        # Clear cells
        del cells[:]

    return errors



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from ..registry import check
from .constraints_checks import create_check_constraint


# Module API

@check('maximum-length-constraint')
def maximum_length_constraint(cells):
    return _check_constraint(cells)


_check_constraint = create_check_constraint('maximum-length-constraint', 'maxLength')



# -*- coding: utf-8 -*-
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

import six
import datetime
import operator
import tabulator
import tableschema
from . import cells
from copy import copy
from multiprocessing.pool import ThreadPool
from .registry import registry
from .error import Error
from . import exceptions
from . import config


# Module API

class Inspector(object):

    # Public

    def __init__(self,
                 checks=['structure', 'schema'],
                 skip_checks=[],
                 infer_schema=False,
                 infer_fields=False,
                 order_fields=False,
                 error_limit=config.DEFAULT_ERROR_LIMIT,
                 table_limit=config.DEFAULT_TABLE_LIMIT,
                 row_limit=config.DEFAULT_ROW_LIMIT):
        """https://github.com/frictionlessdata/goodtables-py#inspector
        """

        # Set attributes
        self.__checks = checks
        self.__skip_checks = skip_checks
        self.__infer_schema = infer_schema
        self.__infer_fields = infer_fields
        self.__order_fields = order_fields

        parse_limit = lambda num: float('inf') if (num < 0) else num  # noqa:E731
        self.__error_limit = parse_limit(error_limit)
        self.__table_limit = parse_limit(table_limit)
        self.__row_limit = parse_limit(row_limit)

    def inspect(self, source, preset=None, **options):
        """https://github.com/frictionlessdata/goodtables-py#inspector
        """

        # Start timer
        start = datetime.datetime.now()

        # Prepare preset
        preset = self.__get_source_preset(source, preset)
        if preset == 'nested':
            options['presets'] = self.__presets
            for s in source:
                if s.get('preset') is None:
                    s['preset'] = self.__get_source_preset(s['source'])

        # Prepare tables
        preset_func = self.__get_preset(preset)['func']
        warnings, tables = preset_func(source, **options)
        if len(tables) > self.__table_limit:
            warnings.append(
                'Dataset inspection has reached %s table(s) limit' %
                (self.__table_limit))
            tables = tables[:self.__table_limit]

        # Collect table reports
        table_reports = []
        if tables:
            tasks = []
            pool = ThreadPool(processes=len(tables))
            try:
                for table in tables:
                    tasks.append(pool.apply_async(self.__inspect_table, (table,)))
                for task in tasks:
                    table_warnings, table_report = task.get()
                    warnings.extend(table_warnings)
                    table_reports.append(table_report)
            finally:
                pool.terminate()

        # Stop timer
        stop = datetime.datetime.now()

        # Compose report
        report = {
            'time': round((stop - start).total_seconds(), 3),
            'valid': all(item['valid'] for item in table_reports),
            'error-count': sum(len(item['errors']) for item in table_reports),
            'table-count': len(tables),
            'tables': table_reports,
            'warnings': warnings,
            'preset': preset,
        }

        return report

    # Internal

    @property
    def __presets(self):
        return registry.compile_presets()

    def __get_preset(self, preset):
        try:
            return self.__presets[preset]
        except KeyError:
            message = 'Preset "%s" is not registered' % preset
            raise exceptions.GoodtablesException(message)

    def __get_source_preset(self, source, preset=None):
        if preset is None:
            preset = 'table'
            if isinstance(source, six.string_types):
                source_path = source.lower()
                if source_path.endswith('datapackage.json') \
                   or source_path.endswith('.zip'):
                    preset = 'datapackage'
            elif isinstance(source, dict):
                if 'resources' in source:
                    preset = 'datapackage'
            elif isinstance(source, list):
                if source and isinstance(source[0], dict) and 'source' in source[0]:
                    preset = 'nested'

        return preset

    def __infer(self, sample, headers=None):
        if headers is None:
            headers = [None] * len(sample[0]) if sample else []
        schema = tableschema.Schema()
        schema.infer(sample, headers=headers)
        return schema

    def __inspect_table(self, table):

        # Start timer
        start = datetime.datetime.now()

        # Prepare vars
        errors = []
        warnings = []
        headers = []
        row_number = 0
        fatal_error = False
        source = table['source']
        stream = table['stream']
        schema = table['schema']
        extra = table['extra']

        # Prepare checks
        checks = registry.compile_checks(
            table.get('checks', self.__checks), self.__skip_checks,
            order_fields=self.__order_fields, infer_fields=self.__infer_fields)

        # Prepare table
        try:
            stream.open()
            sample = stream.sample
            headers = stream.headers
            if headers is None:
                headers = [None] * len(sample[0]) if sample else []
            if _filter_checks(checks, type='schema'):
                if schema is None and self.__infer_schema:
                    schema = self.__infer(sample, headers)
            if schema is None:
                checks = _filter_checks(checks, type='schema', inverse=True)
        except Exception as exception:
            fatal_error = True
            error = _compose_error_from_exception(exception)
            errors.append(error)

        # Prepare schema
        if not fatal_error:
            if schema:
                if schema.primary_key:
                    for field in schema.descriptor.get('fields', []):
                        if field.get('name') in schema.primary_key:
                            field['primaryKey'] = True
                    schema.commit()
                for error in schema.errors:
                    fatal_error = True
                    error = _compose_error_from_schema_error(error)
                    errors.append(error)

        # Head checks
        if not fatal_error:
            # Prepare cells
            fields = [None] * len(headers)
            if schema is not None:
                fields = schema.fields
            header_cells = cells.create_cells(headers, fields)

            has_headers = (None not in headers)
            if has_headers:
                head_checks = _filter_checks(checks, context='head')
                for check in head_checks:
                    if not header_cells:
                        break
                    check_func = getattr(check['func'], 'check_headers', check['func'])
                    errors += (check_func(header_cells, sample) or [])
                for error in errors:
                    error.row = None

            # Grab fields from cells, as there might have been new ones
            # inferred by the "extra-header" check
            fields = [c.get('field') for c in header_cells]

        # Body checks
        if not fatal_error:
            body_checks = _filter_checks(checks, context='body')
            with stream:
                extended_rows = stream.iter(extended=True)
                while True:
                    try:
                        row_number, _, row = next(extended_rows)
                    except StopIteration:
                        break
                    except Exception as exception:
                        fatal_error = True
                        error = _compose_error_from_exception(exception)
                        errors.append(error)
                        break
                    row_cells = cells.create_cells(headers, fields, row, row_number)
                    for check in body_checks:
                        if not row_cells:
                            break
                        check_func = getattr(check['func'], 'check_row', check['func'])
                        errors += (check_func(row_cells) or [])
                    for error in reversed(errors):
                        if error.row is not None:
                            break
                        error.row = row
                    if row_number >= self.__row_limit:
                        warnings.append(
                            'Table "%s" inspection has reached %s row(s) limit' %
                            (source, self.__row_limit))
                        break
                    if len(errors) >= self.__error_limit:
                        warnings.append(
                            'Table "%s" inspection has reached %s error(s) limit' %
                            (source, self.__error_limit))
                        break

        # Table checks
        if not fatal_error:
            for check in checks:
                check_func = getattr(check['func'], 'check_table', None)
                if check_func:
                    errors += (check_func() or [])

        # Stop timer
        stop = datetime.datetime.now()

        # Compose report
        headers = headers if None not in headers else None
        if self.__error_limit != float('inf'):
            errors = errors[:self.__error_limit]
        errors = sorted(errors)
        report = copy(extra)
        report.update({
            'time': round((stop - start).total_seconds(), 3),
            'valid': not bool(errors),
            'error-count': len(errors),
            'row-count': row_number,
            'source': source,
            'headers': headers,
            'scheme': stream.scheme,
            'format': stream.format,
            'encoding': stream.encoding,
            'schema': 'table-schema' if schema else None,
            'errors': [dict(error) for error in errors],
        })

        return warnings, _clean_empty(report)


# Internal

def _filter_checks(checks, type=None, context=None, inverse=False):
    result = []
    comparator = operator.eq if inverse else operator.ne
    for check in checks:
        if type and comparator(check['type'], type):
            continue
        if context and comparator(check['context'], context):
            continue
        result.append(check)
    return result


def _compose_error_from_exception(exception):
    code = 'source-error'
    message = str(exception)

    if isinstance(exception, tabulator.exceptions.SourceError):
        code = 'source-error'
    elif isinstance(exception, tabulator.exceptions.SchemeError):
        code = 'scheme-error'
    elif isinstance(exception, tabulator.exceptions.FormatError):
        code = 'format-error'
    elif isinstance(exception, tabulator.exceptions.EncodingError):
        code = 'encoding-error'
    elif isinstance(exception, tabulator.exceptions.IOError):
        code = 'io-error'
    elif isinstance(exception, tabulator.exceptions.HTTPError):
        code = 'http-error'

    return Error(code, message=message)


def _compose_error_from_schema_error(error):
    message_substitutions = {
        'error_message': error,
    }
    return Error(
        'schema-error',
        message_substitutions=message_substitutions
    )


def _clean_empty(d):
    """Remove None values from a dict."""
    if not isinstance(d, (dict, list)):
        return d
    if isinstance(d, list):
        return [v for v in (_clean_empty(v) for v in d) if v is not None]
    return {
        k: v for k, v in
        ((k, _clean_empty(v)) for k, v in d.items())
        if v is not None
    }
